//----------------------------------------------------------------------------------------
// UDocumentJavelin.cp
// Copyright 1991-1992 Apple Computer, Inc.
//----------------------------------------------------------------------------------------

#ifndef __UDOCUMENTJAVELIN__
#include "UDocumentJavelin.h" 
#endif

#ifndef __UCOMMANDJAVELIN__
#include "UCommandJavelin.h"
#endif

#ifndef __UJAVELINVIEW__
#include "UJavelinView.h"
#endif

#ifndef __UREMOTECOMMANDS__
#include "URemoteCommands.h"
#endif

#ifndef __AUXAESUITE__
#include "AUXAESuite.h"
#endif

#ifndef __UJAVELINDIALOGS__
#include "UJavelinDialogs.h"
#endif

#ifndef __UPRINTING__
#include <UPrinting.h>
#endif

#ifndef TOOLUTILS
#include "ToolUtils.h"
#endif

#ifndef PACKAGES
#include <Packages.h>
#endif


//constants
const short kDefaultSize = 40;
const OSType kDefaultName = 'newV';

//========================================================================================
// CLASS TDocumentJavelin
//========================================================================================


//----------------------------------------------------------------------------------------
// TDocumentJavelin::Initialize: 
//----------------------------------------------------------------------------------------
#pragma segment AOpen
pascal void TDocumentJavelin::Initialize() // Override 
{
	inherited::Initialize();
	
	fAEAddressDesc.descriptorType = typeNull;
	fAEAddressDesc.dataHandle = NULL;
	
	fJavelinView 		= NULL;
	fCommandOut 		= FALSE;
	fWindowTitle		= gEmptyString;
	fHeartBeatCounter	= TickCount()/60;    //gotta do it here because DoIdle waits for nobody!
	fPortRefNum			= NULL;
	fHostVersion		= gEmptyString;
	fInterval			= 60;
}

//----------------------------------------------------------------------------------------
// TDocumentJavelin::IDocumentJavelin: 
//----------------------------------------------------------------------------------------
#pragma segment AOpen
pascal void TDocumentJavelin::IDocumentJavelin()
{
	this->IDocument();

} 

//----------------------------------------------------------------------------------------
// TDocumentJavelin::Close: 
//----------------------------------------------------------------------------------------
pascal void TDocumentJavelin::Disconnect()
{
	this->fAEAddressDesc.descriptorType = typeNull;
	this->fAEAddressDesc.dataHandle = NULL;
	

} // TDocumentJavelin::Close 

//----------------------------------------------------------------------------------------
// TDocumentJavelin::Close: 
//----------------------------------------------------------------------------------------
pascal void TDocumentJavelin::Close()
{
	CWindowIterator iter(this);

	for (TWindow* aWindow = iter.FirstWindow(); iter.More(); aWindow = iter.NextWindow())
		aWindow->CloseAndFree();
	
	if (Connected())
		{
		TQuitCmd* aQuitCmd = new TQuitCmd;
		aQuitCmd->IQuitCmd(this);
		aQuitCmd->DoIt();    //otherwise the doc is gone if I post it and kaboom!

		this->Disconnect();
		}
} // TDocumentJavelin::Close 

//----------------------------------------------------------------------------------------
// TDocumentJavelin::Free: 
//----------------------------------------------------------------------------------------
#pragma segment AClose
pascal void TDocumentJavelin::Free() // Override 
{
	AEAddressDesc aDesc = fAEAddressDesc;
	AEDisposeDesc(aDesc);

	inherited::Free();
}
//----------------------------------------------------------------------------------------
// TDocumentJavelin::SetHostVersion: 
//----------------------------------------------------------------------------------------
#pragma segment AClose
pascal void TDocumentJavelin::SetHostVersion(CStr255 theVersion)
{
	fHostVersion = theVersion;
}

//----------------------------------------------------------------------------------------
// TDocumentJavelin::DoMakeViews: 
//----------------------------------------------------------------------------------------
#pragma segment AOpen
pascal void TDocumentJavelin::DoMakeViews(Boolean /*forPrinting*/) // Override 
{
	TWindow* aWindow = NULL;
	TStdPrintHandler* aHandler = NULL;
	TJavelinView* aJavelinView = NULL;
	CPoint resize;
	resize.v = 20;
	resize.h = 496;

	FailNIL(aWindow = gViewServer->NewTemplateWindow(kJavelinWindowID, this));
//	aWindow->Zoom(inZoomOut);
	aWindow->SetResizeLimits(resize,resize);

	aJavelinView = (TJavelinView*) (aWindow->FindSubView('view'));	// Must cast because FindSubView returns TView
	fJavelinView = aJavelinView;

	this->DoMenuCommand(cGetPPCServerCmd);      // launches ppc browser on startup  (actually on new document)

	if (!this->Connected())
		Failure(0,0);

	
	aHandler = new TStdPrintHandler;
	aHandler->IStdPrintHandler(this,			// its document 
							   aJavelinView,		// its view 
							   !kSquareDots,	// does not have square dots 
							   kFixedSize,		// horzontal page size is fixed 
							   kFixedSize);		// vertical page size is fixed 

	fHeartBeatCounter	= TickCount()/60;
} // TDocumentJavelin::DoMakeViews 

//----------------------------------------------------------------------------------------
// TDocumentJavelin::DoMenuCommand: This method is overridden to handle menu items which
// are enabled when this document is in the target chain. In this example, this is true
// when the document is open and its window is the active window. The inherited method
// should always be called so that MacApp can allow successor objects in the target chain
// (i.e. the application) to handle THEIR menu items.
//----------------------------------------------------------------------------------------
#pragma segment ASelCommand
pascal void TDocumentJavelin::DoMenuCommand(CommandNumber aCommandNumber) // Override 
{
	
	switch (aCommandNumber) 
	{
		case cGetPPCServerCmd: 
			TLookupLauncher* aCommand = new TLookupLauncher;
			aCommand->ILookupLauncher(aCommandNumber,this);
			aCommand->DoIt();
			//this->PostCommand(aCommand);
			break;
		case cConfig1:
			fInterval = 1;
			TIntervalCmd* aCommand1 = new TIntervalCmd;
			aCommand1->IIntervalCmd(this);
			this->PostCommand(aCommand1);
			break;
		case cConfig10: 
			fInterval = 10;
			TIntervalCmd* aCommand2 = new TIntervalCmd;
			aCommand2->IIntervalCmd(this);
			this->PostCommand(aCommand2);
			break;
		case cConfig30: 
			fInterval = 30;
			TIntervalCmd* aCommand3 = new TIntervalCmd;
			aCommand3->IIntervalCmd(this);
			this->PostCommand(aCommand3);
			break;
		case cConfig60: 
			fInterval = 60;
			TIntervalCmd* aCommand4 = new TIntervalCmd;
			aCommand4->IIntervalCmd(this);
			this->PostCommand(aCommand4);
			break;
		default:
			inherited::DoMenuCommand(aCommandNumber);
			break;
	}
} 


//----------------------------------------------------------------------------------------
// TDocumentJavelin::DoSetupMenus: This method is overridden to enable menu items which
// should be enabled when this object is in the target chain. MacApp initially disables
// all menu items, then lets the objects in the target chain enable those items they
// handle.
//
// A document object is in the target chain when its window is the active window.
//
// The inherited method is called so that TDocument can enable document-level menu items
// like "Save ". This also ensures that objects further up the target chain (the
// application) can set up THEIR menus.
//----------------------------------------------------------------------------------------
#pragma segment ARes
pascal void TDocumentJavelin::DoSetupMenus() // Override 
{
	inherited::DoSetupMenus();
	
	if (this->Connected())
		{
		EnableCheck(cConfig1,TRUE,this->fInterval == 1);
		EnableCheck(cConfig10,TRUE,this->fInterval == 10);
		EnableCheck(cConfig30,TRUE,this->fInterval == 30);
		EnableCheck(cConfig60,TRUE,this->fInterval ==60);
		
		}
} 
//----------------------------------------------------------------------------------------
#pragma segment ARes
pascal void TDocumentJavelin::DoPostMakeViews(Boolean forPrinting)
{
	if (Connected())
	{
		this->DoMenuCommand(cConfig30);
		this->SetTitle(this->fWindowTitle);
		inherited::DoPostMakeViews(forPrinting);
	
	}
	else
		this->Close();
		
fHeartBeatCounter	= TickCount()/60;
} // TDocument::DoPostMakeViews 

//----------------------------------------------------------------------------------------
// TDocumentJavelin::GetAEAddressDesc: 
//----------------------------------------------------------------------------------------

#pragma segment ADoCommand
pascal void TDocumentJavelin::GetAEAddressDesc(AEAddressDesc& theAddress) 
{
	theAddress = fAEAddressDesc;
}
//----------------------------------------------------------------------------------------
// TDocumentJavelin::Connected: 
//----------------------------------------------------------------------------------------

#pragma segment ARes
pascal Boolean TDocumentJavelin::Connected() 
{
	return (fAEAddressDesc.dataHandle != NULL);
}
//----------------------------------------------------------------------------------------
// TDocumentJavelin::Connected: 
//----------------------------------------------------------------------------------------

#pragma segment ARes
pascal void TDocumentJavelin::HaveAHeartBeat() 
{
	fHeartBeatCounter = TickCount()/60;
}
//----------------------------------------------------------------------------------------
// TDocumentJavelin::CheckPulse: 
//----------------------------------------------------------------------------------------

#pragma segment ARes
pascal Boolean TDocumentJavelin::CheckPulse() 
{
	long now = TickCount()/60;
	
	if (((now - this->fHeartBeatCounter) > 120) && (!this->fWindowTitle.IsEmpty()) )      //2 minutes
	{

		CStr255 theOtherString	;
		
		theOtherString = this->fWindowTitle;
		this->fWindowTitle	= gEmptyString;      //this is my flag to let me know that I have one posed already
		
		JavelinErrorBoxNumber(kCheckPulseError, theOtherString);
		
		return FALSE;   //this doc is a gonner 
	}
	else if ((now - this->fHeartBeatCounter) > 60  &&  (!this->fWindowTitle.IsEmpty()))      //1 minute
		{
		SysBeep(2);
		return TRUE;    //this doc is still ok
		}
	return TRUE;        //steady as she goes!
}
//----------------------------------------------------------------------------------------
// TDocumentJavelin::PendingCommand: 
//----------------------------------------------------------------------------------------
#pragma segment ARes
pascal void TDocumentJavelin::PendingCommand()
{
	if (!fCommandOut) {
		fCommandOut = TRUE;
		}
		
}

//----------------------------------------------------------------------------------------
// TDocumentJavelin::CommandCompleted: 
//----------------------------------------------------------------------------------------
#pragma segment ARes
pascal void TDocumentJavelin::CommandCompleted()
{
	if (fCommandOut) {

		fCommandOut = FALSE;   // done is done right?
		
	}
	
}

//----------------------------------------------------------------------------------------
// TDocumentJavelin::HasPendingCommands: 
//----------------------------------------------------------------------------------------
#pragma segment ARes
pascal Boolean TDocumentJavelin::HasPendingCommands()
{
	return fCommandOut;
}

//----------------------------------------------------------------------------------------
// TDocumentJavelin::SetAddress: 
//----------------------------------------------------------------------------------------
#pragma segment ADoCommand
pascal void TDocumentJavelin::SetAddress(TargetID& theTarget) 
{
	// Accepts result of PPCBrowser call. May fail.

	AEAddressDesc	targetAddress;

	FailOSErr(AECreateDesc(typeTargetID, Ptr(&theTarget), sizeof(theTarget),
						   targetAddress));
	fAEAddressDesc = targetAddress;
}
//----------------------------------------------------------------------------------------
// TDocumentJavelin::SetPortRefNum: 
//----------------------------------------------------------------------------------------
#pragma segment ADoCommand
pascal void TDocumentJavelin::SetPortRefNum(short thePortRefNum) 
{
	fPortRefNum = thePortRefNum;
}
//----------------------------------------------------------------------------------------
// TDocumentJavelin::GetPortRefNum: 
//----------------------------------------------------------------------------------------
#pragma segment ADoCommand
pascal short TDocumentJavelin::GetPortRefNum() 
{
	return fPortRefNum;
}

//----------------------------------------------------------------------------------------
#pragma segment ADoCommand
pascal void TDocumentJavelin::SetTitle(const CStr255& aTitle)
{
	inherited::SetTitle(aTitle);
	((TWindow*)fJavelinView->fSuperView)->SetTitle(aTitle);
}
//----------------------------------------------------------------------------------------
#pragma segment ADoCommand
pascal void TDocumentJavelin::SetfWindowTitle(const CStr255& aTitle)
{
	this->fWindowTitle = aTitle;
}

//----------------------------------------------------------------------------------------
#pragma segment AVGRes
pascal Boolean TDocumentJavelin::LocateDocument(long theSessionID)
{
	return fPortRefNum == theSessionID;
}

//--------------------------------------------------------------------------------------------------
// TDocumentJavelin::SetWindowText: 
//----------------------------------------------------------------------------------------
#pragma segment ARes
pascal void TDocumentJavelin::SetWindowText(CStr255 theText)
{
	if (fJavelinView)
		{
		TStaticText* theStatText  =  (TStaticText*) fJavelinView->FindSubView('stat');
		FailNIL(theStatText);
		theStatText->SetText(theText, TRUE);
		//theStatText->DrawContents();
		}
}




